import os
import plistlib

try:
    import bsdiff4
except ImportError:
    raise


class Bundle:
    def __init__(self, bundle: str):
        super().__init__()

        self.bundle = bundle

    def checkBundle(self):
        try:
            os.path.isdir(self.bundle)
        except ValueError:
            print('{} is not a bundle!'.format(self.bundle))
            raise

    def parsePlist(self):
        with open('{}/Info.plist'.format(self.bundle), 'rb') as f:
            data = plistlib.load(f)
            return data

    def patch(self, path: str, patchfile: str):
        print('Patching: {}'.format(path))
        bsdiff4.file_patch(path, '{}.patched'.format(
            os.path.basename(path)), patchfile)



try:
    import usb.core
    import usb.util
except ImportError:
    print('Failed to import PyUSB!')
    raise


class USB:
    def __init__(self):
        super().__init__()

        # Some stuff with WTF, maybe we can force DFU somehow?

        # 05ac:1260 Nano 2G regular/disk mode

        # ffff:8642 Nano 2G iBugger Loader

        # dev.idVendor == 0x05ac and dev.idProduct == 0x1220 Bootrom DFU

        # dev.idVendor == 0x05ac and dev.idProduct == 0x1240 Nano 2G NOR DFU (WTF I think/SoftDFU)

        # 05ac:12a8 Normal mode iPhone 5S

        # 05ac:1227 DFU mode iPhone 5S

        # 05ac:1281 Recovery mode iPhone 5S

        # Uploading /home/merculous/Documents/ibugger/ibugger/core-2.bin to 0x22000000.... done
        # Passing control to code at 0x22000020... done
        # Connected to iBugger Core v0.1.1 on iPod Nano 2G, USB version 00.01

        # TODO Add Pwnage 2.0 support. Just uploading a malformed certificate/file


    def getDevices(self):
        arm64_normal_mode = usb.core.find(idVendor=0x05ac, idProduct=0x12a8)

    def sendPacket(self):
        pass

    def unpackPacket(self, packet):
        pass

    def debugNano2G(self):
        pass

    def sendCommand(self, command):
        pass

    def enterRecovery(self):
        pass

    def reboot(self):
        pass



import plistlib

try:
    from .utils import getDeviceType, getMajorDeviceRevision, getMinorDeviceRevision, fastTokenHex
except ImportError:
    pass


class BuildManifest(object):
    def __init__(self, data: str):
        super().__init__()

        try:
            stuff = plistlib.loads(data)
        except IOError:
            print('Failed to read manifest!')
            raise
        else:
            self.data = stuff

    def parse(self):
        data = self.data
        return data


class TSSManifest(object):
    def __init__(self):
        super().__init__()

    # Thanks tihmstar! http://blog.tihmstar.net/2017/01/basebandgoldcertid-not-found-please.html
    # https://github.com/tihmstar/tsschecker/blob/master/tsschecker/tsschecker.c#L110

    #  Returns BbGoldCertId, BbSNUM length

    def getBbConfigurationForDevice(self, device: str):
        if getDeviceType(device) == 'iPhone':
            if getMajorDeviceRevision(device) == 1 \
                    or getMajorDeviceRevision(device) == 2:
                return []

            if getMajorDeviceRevision(device) == 3:
                if getMinorDeviceRevision(device) <= 2:
                    return [257, 12]
                else:
                    return [2, 4]

            if getMajorDeviceRevision(device) == 4:
                return [2, 4]

            if getMajorDeviceRevision(device) == 5:
                if getMinorDeviceRevision(device) <= 2:
                    return [3255536192, 4]
                else:
                    return [3554301762, 4]

            if getMajorDeviceRevision(device) == 6:
                return [3554301762, 4]

            if getMajorDeviceRevision(device) == 7:
                return [3840149528, 4]

            if getMajorDeviceRevision(device) == 8:
                return [3840149528, 4]

            if getMajorDeviceRevision(device) == 9:
                if getMinorDeviceRevision(device) <= 2:
                    return [2315222105, 4]
                else:
                    return [1421084145, 12]

            if getMajorDeviceRevision(device) == 10:
                if getMinorDeviceRevision(device) <= 3:
                    return [2315222105, 4]
                else:
                    return [524245983, 12]

            if getMajorDeviceRevision(device) == 11:
                return [165673526, 12]

            if getMajorDeviceRevision(device) == 12:
                return [524245983, 12]

        if getDeviceType(device) == 'iPad':
            if getMajorDeviceRevision(device) == 1:
                return []

            if getMajorDeviceRevision(device) == 2:
                if getMinorDeviceRevision(device) >= 2 \
                        and getMinorDeviceRevision(device) <= 3:
                    return [257, 12]

                if getMinorDeviceRevision(device) >= 6 \
                        and getMinorDeviceRevision(device) <= 7:
                    return [3255536192, 4]

            if getMajorDeviceRevision(device) == 3:
                if getMinorDeviceRevision(device) >= 2 \
                        and getMinorDeviceRevision(device) <= 3:
                    return [4, 4]

                if getMinorDeviceRevision(device) >= 5:
                    return [3255536192, 4]

            if getMajorDeviceRevision(device) == 4:
                if (getMinorDeviceRevision(device) >= 2
                    and getMinorDeviceRevision(device) <= 3) \
                                                         \
                    or (getMinorDeviceRevision(device) >= 5
                        and getMinorDeviceRevision(device) <= 6) \
                                                         \
                    or (getMinorDeviceRevision(device) >= 8
                        and getMinorDeviceRevision(device) <= 9):
                    return [3554301762, 4]

            if getMajorDeviceRevision(device) == 5:
                if getMinorDeviceRevision(device) == 2 \
                        or getMinorDeviceRevision(device) == 4:
                    return [3840149528, 4]

            if getMajorDeviceRevision(device) == 6:
                if getMinorDeviceRevision(device) == 4 \
                        or getMinorDeviceRevision(device) == 8 \
                        or getMinorDeviceRevision(device) == 11:
                    return [3840149528, 4]

            if getMajorDeviceRevision(device) == 7:
                if getMinorDeviceRevision(device) == 2 \
                        or getMinorDeviceRevision(device) == 4:
                    return [2315222105, 4]

                if getMinorDeviceRevision(device) == 6:
                    return [3840149528, 4]

                if getMinorDeviceRevision(device) == 12:
                    return [524245983, 12]

            if getMajorDeviceRevision(device) == 8:
                if (getMinorDeviceRevision(device) >= 3
                    and getMinorDeviceRevision(device) <= 4) \
                                                         \
                    or (getMinorDeviceRevision(device) >= 7
                        and getMinorDeviceRevision(device) <= 8):
                    return [165673526, 12]

            if getMajorDeviceRevision(device) == 11:
                if getMinorDeviceRevision(device) == 2 \
                        or getMinorDeviceRevision(device) == 4:
                    return [165673526, 12]

        return []

    # See 'Notes' in https://www.theiphonewiki.com/wiki/SHSH_Protocol#Communication
    def createTSSTestVersionManifest(self, path: str):
        testVersionManifest = dict(
            ApSecurityDomain=1
        )

        with open(path, 'wb+') as v:
            # I'm getting module plistlib has no 'FMT_XML' member error from pylint, according to docs, its supported, guess this is a bug?
            plistlib.dump(testVersionManifest, v, fmt=plistlib.FMT_XML)

    # See 'Sending data (request)' in https://www.theiphonewiki.com/wiki/SHSH_Protocol#Communication
    def initFromBuildManifest(self, device: str, tss_manifest_path: str, build_manifest_path: str, ecid: str, apnonce='', sepnonce='', bbsnum=''):
        with open(build_manifest_path, 'rb+') as f:
            data = plistlib.load(f, fmt=plistlib.FMT_XML)

            has_baseband = False
            found_erase_identity = False

            # Set TSS manifest to the Erase preset
            for identity in data['BuildIdentities']:
                if 'Erase' in identity['Info']['Variant']:
                    for component in identity['Info']['VariantContents']:
                        if component == 'SEP':  # Set SepNonce now
                            if sepnonce == '':
                                sepnonce = fastTokenHex(20)

                            identity['SepNonce'] = int(
                                sepnonce, 16).to_bytes(20, 'big')

                        elif component == 'BasebandFirmware':  # Enable Baseband-related keys later
                            has_baseband = True

                    # This will clash with the root key labeled 'Info'
                    del identity['Info']

                    data = identity
                    found_erase_identity = True
                    break

            if not found_erase_identity:
                print('No \'Erase\' BuildIdentity was found in the BuildManifest')
                exit(1)

            for key in ['Info', 'ProductMarketingVersion']:  # Remove unnecessary root keys
                try:
                    del data[key]
                except:
                    pass

            # production_mode = False
            # security_mode = False

            for component in data['Manifest']:
                content = data['Manifest'][component]

                #  This is messy; fix later

                if 'RestoreRequestRules' in content['Info']:
                    for rule in content['Info']['RestoreRequestRules']:
                        if 'EPRO' in rule['Actions']:
                            content['EPRO'] = True
                            if 'ApRequiresImage4' in rule['Conditions']:
                                if rule['Conditions']['ApRequiresImage4']:
                                    data['@ApImg4Ticket'] = True

                        elif 'ESEC' in rule['Actions']:
                            content['ESEC'] = True

                if 'Trusted' in content and 'Digest' not in content:
                    content['Digest'] = bytes()

                del content['Info']

                data[component] = content  # Move keys to root

            if '@ApImg4Ticket' not in data:
                data['@APTicket'] = True  # Old IMG3 ticket

            del data['Manifest']

            #  These are hardcoded for now; will change later

            data['ApProductionMode'] = True
            data['ApSecurityMode'] = True

            if has_baseband:
                bb_config = self.getBbConfigurationForDevice(device)
                if bb_config != [] and bbsnum:
                    if 'BbChipID' in data:
                        data['BbChipID'] = int(data['BbChipID'], 16)

                    data['@BBTicket'] = True
                    data['BbGoldCertId'] = bb_config[0]

                    if len(bbsnum) != bb_config[1]:
                        print('Provided BbSNUM length is', len(bbsnum),
                              'while the max for', device, 'is', bb_config[1])
                        exit(1)

                    data['BbSNUM'] = int(bbsnum, 16).to_bytes(
                        bb_config[1], 'big')
                else:
                    print(
                        device, 'either does not have a known BbGoldCertId and/or you have not supplied a BbSNUM. Skipping BBTicket...')
                    for key in list(data):
                        if key == 'BasebandFirmware' or key[0:2] == 'Bb':
                            del data[key]

            if len(ecid) == 13 or len(ecid) == 11:
                try:
                    data['ApECID'] = int(ecid, 16)  # Hex ECID
                except:
                    data['ApECID'] = int(ecid)      # Decimal ECID
            else:
                data['ApECID'] = int(ecid)          # Decimal ECID

            if apnonce == '':
                apnonce = fastTokenHex(20)

            data['ApNonce'] = int(apnonce, 16).to_bytes(20, 'big')

        with open(tss_manifest_path, 'wb+') as p:
            plistlib.dump(data, p, fmt=plistlib.FMT_XML)

        return {
            'apnonce':  apnonce,
            'sepnonce': sepnonce
        }



import re
# import plistlib

# try:
# from .ipwapi import APIParser
# from .manifest import BuildManifest
# except ImportError:
# raise

"""

This will handle the key template stuff from iphonewiki (webpage)

"""

"""

This is the key template for the key pages. When a page uses this template, it is added to the Key Pages category.

Syntax:

If the device is not an iPhone or a cellular iPad, delete the Baseband parameter.

If the device is an Apple TV (iOS versions), set Version to the marketing version with the internal version in parenthesis.

For each *, *IV, and *Key section, put the file name on the first line. If the item is the Root FS or a ramdisk, remove the .dmg
file extension. If that firmware item does not exist in that firmware, delete its parameter section. If the firmware item is not encrypted,
set *IV to "Not Encrypted" and delete the *Key parameter. If the key/iv is not known, then the KBAG should be added.

For devices such as iPhone 6s that have two sets of files such as N71AP and N71mAP, then a second file can be added underneath
with a "2" at the end of the name. For example, you can have AppleLogo and AppleLogo2 blocks.

For DownloadURL, do not place anything other than URLs to free firmwares hosted on Apple Inc.'s servers. For beta firmwares, delete the parameter.

"""


class Template(object):
    def __init__(self):
        super().__init__()

    def createTemplate(self):
        # TODO Make this, completely scrap the template file I have, prepare for pypi functionality
        # manifest = BuildManifest()
        # data = manifest.extractData()

        # api = APIParser(data['device'], data['ios'])
        # url = api.printURLForArchive()

        # raw_template = [
        # "{{keys",
        # "| Version              = ".format(data['ios']),
        # "| Build                = ".format(data['buildid']),
        # "| Device               = ".format(data['device']),
        # "| Codename             = ".format(data['codename']),
        # "| Baseband             = ",
        # "| DownloadURL          = ".format(url),
        # "| RootFS               = ",
        # "| RootFSKey            = ",
        # "| {}                   = ",
        # "| {}IV                 = ",
        # "| {}Key                = ",
        # "| {}KBAG               = ",
        # "}}"
        # ]
        pass

    # FIXME I know I can make this a little better...

    def parseTemplate(self, data: str):
        keys = data.split('{{keys')[1].split('}}')[0].splitlines()
        new_list = list(filter(None, keys))  # Remove all ''
        fixed = list()
        for stuff in new_list:
            fix = re.sub(r'\s+', ' ', stuff).strip().replace('| ', '')
            epic_fix = fix.replace(' =', '').split()
            thingy = {epic_fix[0]: epic_fix[1]}
            fixed.append(thingy)
        return fixed



import sys
from random import choice, seed
from string import ascii_letters, digits

try:
    import progressbar
except ImportError:
    raise

"""

All of the helper functions or just a module to store other functions
that don't have a particular module that its similar to.

Basically just 'tools'.

"""

# Maybe convert progress into my own custom file downloader that auto grabs the data such as filesize, duration, etc.

# https://stackoverflow.com/questions/37748105/how-to-use-progressbar-module-with-urlretrieve

pbar = None


def showProgress(block_num, block_size, total_size):
    global pbar
    if pbar is None:
        pbar = progressbar.ProgressBar(maxval=total_size)
        pbar.start()
    downloaded = block_num * block_size
    if downloaded < total_size:
        pbar.update(downloaded)
    else:
        pbar.finish()
        pbar = None


def splitKbag(kbag: str):
    if len(kbag) != 96:
        sys.exit('String provided is not 96 bytes! The length read was:', len(kbag))

    iv = kbag[:32]
    key = kbag[-64:]
    data = {
        'iv': iv,
        'key': key
    }
    return data


def getDeviceType(device: str):
    for index in range(0, len(device)-3):
        if device[index] in digits:
            return device[:index]
    return False


def getMajorDeviceRevision(device: str):
    splitting_point = device.find(',')
    for index in range(splitting_point-1, 3, -1):
        if device[index] in ascii_letters:
            return int(device[index+1:splitting_point])
    return -1


def getMinorDeviceRevision(device: str):
    return int(device[device.find(',')+1:])


def fastTokenHex(byte_length: int):
    hexdigits = '0123456789abcdef'
    token = ''

    seed()
    for i in range(0, byte_length * 2):
        token += choice(hexdigits)
    return token



import struct

'''

typedef struct img3File {
    uint32_t magic;       // ASCII_LE("Img3")
    uint32_t fullSize;    // full size of fw image
    uint32_t sizeNoPack;  // size of fw image without header
    uint32_t sigCheckArea;// although that is just my name for it, this is the
                          // size of the start of the data section (the code) up to
                          // the start of the RSA signature (SHSH section)
    uint32_t ident;       // identifier of image, used when bootrom is parsing images
                          // list to find LLB (illb), LLB parsing it to find iBoot (ibot),
                          // etc.
    img3Tag  tags[];      // continues until end of file
};
typedef struct img3Tag {
    uint32_t magic;            // see below
    uint32_t totalLength;      // length of tag including "magic" and these two length values
    uint32_t dataLength;       // length of tag data
    uint8_t  data[dataLength];
    uint8_t  pad[totalLength - dataLength - 12]; // Typically padded to 4 byte multiple
};

'''

'''

VERS: iBoot version of the image
SEPO: Security Epoch
SDOM: Security Domain
PROD: Production Mode
CHIP: Chip to be used with. example: 0x8900 for S5L8900.
BORD: Board to be used with
KBAG: Contains the IV and key required to decrypt; encrypted with the GID Key
SHSH: RSA encrypted SHA1 hash of the file
CERT: Certificate
ECID: Exclusive Chip ID unique to every device
TYPE: Type of image, should contain the same string as the header's ident
DATA: Real content of the file
NONC: Nonce used when file was signed.
CEPO: Chip epoch
OVRD: For demoting -axi0mX
RAND:
SALT: For flavor -axi0mX (lmao)

'''

'''

Decryption is done using the modulus at cert + 0xA15
0xC to SHSH is SHAed

'''


class IMG3:
    def __init__(self, file):
        super().__init__()

        with open(file, 'rb') as f:
            self.data = f.read()
            self.info = {
                'magic': struct.unpack('4s', self.data[0:4])[0][::-1].decode(),
                'size': struct.unpack('I', self.data[4:8])[0],
                'unpacksize': struct.unpack('I', self.data[8:12])[0],
                'sigcheckarea': struct.unpack('I', self.data[12:16])[0],
                'ident': struct.unpack('4s', self.data[16:20])[0][::-1].decode(),
                'tags': None,
                'kbag': None,
                'version': None
            }

            # Add tag values

            tags = list()

            i = 20

            while i < 20 + self.info['unpacksize']:
                self.tag_info = {
                    'magic': struct.unpack('4s', self.data[i:i+4])[0][::-1].decode(),
                    'totalsize': struct.unpack('I', self.data[i+4:i+8])[0],
                    'datasize': struct.unpack('I', self.data[i+8:i+12])[0],
                    'data': None
                }

                tag_size = self.tag_info['totalsize']
                self.tag_info['data'] = self.data[i+12:i+tag_size]

                tags.append(self.tag_info)
                self.info['tags'] = tags

                # Done with tag, move on to next

                i += tag_size

            # Add iBoot version string, if iBoot, of course

            for version in self.info['tags']:
                if version['magic'] == 'VERS':
                    version_info = {
                        'stringsize': struct.unpack('I', version['data'][:4])[0],
                        'string': None
                    }
                    size = version_info['stringsize']
                    version_info['string'] = struct.unpack(
                        '{}s'.format(size), version['data'][4:4+size])[0].decode()
                    self.info['version'] = version_info

            # Add kbag values

            kbags = list()
            for kbag in self.info['tags']:
                if kbag['magic'] == 'KBAG':
                    kbag_info = {
                        'type': struct.unpack('I', kbag['data'][:4])[0],
                        'aes_type': struct.unpack('I', kbag['data'][4:8])[0],
                        'kbag': kbag['data'][8:56].hex()
                    }
                    kbags.append(kbag_info)
            self.info['kbag'] = kbags

            # Add SHSH values

            for blob in self.info['tags']:
                if blob['magic'] == 'SHSH':
                    pass

            # Add CERT values

            for cert in self.info['tags']:
                if cert['magic'] == 'CERT':
                    pass

    def printInfo(self):
        return self.info

    def printTags(self):
        return self.info['tags']

    def getKBAGS(self):
        return self.info['kbag']



#!/usr/bin/env python3

"""

This is basically what the frontend is, if that's even correct terminology.

Handles arguments and passes them to the functions.

"""

import sys
from argparse import ArgumentParser

try:
    # from .bundle import Bundle
    # from .foreman import Foreman
    from .img3 import IMG3
    # from .img4 import IMG4
    from .iphonewiki import iPhoneWiki
    # from .ipsw import IPSW
    from .ipswapi import API
    # from .manifest import BuildManifest
    # from .template import Template
    from .tss import TSS
    from .utils import splitKbag
    # from .usb import USB
except ImportError:
    raise


def main():
    """

    Main file to interact with the ipsw.me api and such.
    Some stuff to parse iOS and etc.

    """

    parser = ArgumentParser(
        usage='iospytools <option> <args>',
        description='provides useful tools/commands which are used in iOS research')

    parser.add_argument(
        "--buildid",
        help="Convert an iOS to its buildid",
        nargs=2,
        metavar=('DEVICE', 'iOS'))

    parser.add_argument(
        "--codename",
        help="Get the codname of an iOS",
        nargs=2,
        metavar=('DEVICE', 'iOS'))

    parser.add_argument(
        "--download",
        help="Download an IPSW",
        nargs=2,
        metavar=('DEVICE', 'iOS'))

    parser.add_argument(
        "--img3",
        help="Print info of an img3",
        nargs=1,
        metavar='FILE'
    )

    parser.add_argument(
        "--keys",
        help="Get keys for an iOS",
        nargs=2,
        metavar=('DEVICE', 'iOS'))

    parser.add_argument(
        "--manifest",
        help="Download a Build Manifest from an iOS",
        nargs=2,
        metavar=('DEVICE', 'iOS'))

    parser.add_argument(
        "--shsh",
        help="Save SHSH for all signed iOS versions",
        nargs=2,
        metavar=('DEVICE', 'ECID'))

    parser.add_argument(
        "--signed",
        help="Print the signed iOS versions for a device",
        nargs=1,
        metavar='DEVICE')

    parser.add_argument(
        "--split",
        help="Split a GID decrypted key",
        nargs=1,
        metavar='KEY')

    parser.add_argument(
        "--tags",
        help="Print the tags of an img3",
        nargs=1,
        metavar='FILE'
    )

    parser.add_argument('--test', action='store_true')

    args = parser.parse_args()

    if args.buildid:  # ./yeet --buildid device iOS
        api = API(args.buildid[0], args.buildid[1])
        print(api.iOSToBuildid())

    elif args.codename:  # ./yeet --codename device iOS
        api = API(args.codename[0], args.codename[1])
        print(api.getCodename())

    elif args.download:  # ./yeet --download device iOS
        api = API(args.download[0], args.download[1])
        api.downloadIPSW()

    elif args.img3:
        info = IMG3(args.img3[0])
        print(info.printInfo())

    elif args.keys:  # ./yeet --keys device iOS
        wiki = iPhoneWiki(args.keys[0], args.keys[1])
        data = wiki.getWikiKeys()
        for stuff in data:
            print(stuff)

    elif args.manifest:  # ./yeet --manifest device iOS
        api = API(args.manifest[0], args.manifest[1])
        api.downloadFileFromArchive('BuildManifest.plist')

    elif args.shsh:  # ./yeet --shsh device ecid
        tss = TSS(args.shsh[0], args.shsh[1])
        tss.saveBlobs()

    elif args.signed:  # ./yeet --signed device
        api = API(args.signed[0])
        for stuff in api.getSignedVersions():
            print('{} ({}) {}'.format(
                stuff['iOS'], stuff['buildid'], stuff['filetype']))

    elif args.split:  # ./yeet --split kbag
        kbag = splitKbag(args.split[0])
        print('IV:', kbag['iv'])
        print('Key:', kbag['key'])

    elif args.tags:
        tags = IMG3(args.tags[0])
        tags.printTags()

    else:
        sys.exit(parser.print_help(sys.stderr))


if __name__ == '__main__':
    main()



import os
import zipfile

try:
    # from .bundle import Bundle
    from .manifest import BuildManifest
except ImportError:
    raise

# TODO Create a write-up to make on how to make these patches manually


class IPSW(object):
    def __init__(self, ipsw: str):
        super().__init__()

        try:
            zipfile.is_zipfile(ipsw)
        except ValueError:
            print('{} is not a zip archive!'.format(ipsw))
            raise
        else:
            self.ipsw = ipsw

    def parse(self):
        with zipfile.ZipFile(self.ipsw, 'r') as f:
            f.extract('BuildManifest.plist')
            data = BuildManifest()
            stuff = data.extractData()
            os.remove('BuildManifest.plist')
            return stuff




import json
import os
import plistlib
import sys
from urllib.request import urlopen, urlretrieve

try:
    from remotezip import RemoteZip
except ImportError:
    raise

try:
    from .utils import showProgress
except ImportError:
    raise


class API(object):
    def __init__(self, device: str, version=False, buildid=False, ota=False, beta=False):
        super().__init__()

        self.device = device
        self.version = version

        if buildid:
            self.buildid = buildid
        else:
            self.buildid = self.iOSToBuildid()

        if ota:
            self.ota = ota

        if beta:
            self.beta = beta

    def getDeviceJSONData(self, ota=False):
        # Returns JSON Data, rather than downloading each time, see below later on :P

        if ota:
            api_url = 'https://api.ipsw.me/v4/device/{}?type={}'.format(
                self.device, 'ota')
        else:
            api_url = 'https://api.ipsw.me/v4/device/{}?type={}'.format(
                self.device, 'ipsw')

        try:
            r = urlopen(api_url).read().decode('utf-8')
        except ConnectionError:
            print('Failed to download JSON data!')
            raise
        else:
            return json.loads(r)

    def iOSToBuildid(self):
        # Easy to use function since we don't have to be paranoid about reading leftover json data :D
        data = self.getDeviceJSONData()
        for i in range(0, len(data['firmwares'])):
            if data['firmwares'][i]['version'] == self.version:
                return data['firmwares'][i]['buildid']

    def getArchiveURL(self, filetype='ipsw'):
        if filetype == 'ipsw':
            api_url = 'https://api.ipsw.me/v4/ipsw/{}/{}'.format(
                self.device, self.iOSToBuildid())
        elif filetype == 'ota':
            # TODO Add ota support
            pass
        else:
            raise ValueError('Only filetypes, "ipsw" and "ota" are supported!')

        r = urlopen(api_url).read().decode('utf-8')
        data = json.loads(r)
        url = data['url']
        return url

    def downloadIPSW(self):
        # TODO Use given hashes to determine if we need to download a genuine copy or not
        url = self.getArchiveURL()
        ipsw_name = os.path.basename(url)
        if not os.path.exists(ipsw_name):
            print('URL: {}'.format(url))
            print('File: {}'.format(ipsw_name))
            urlretrieve(url, ipsw_name, showProgress)
        else:
            sys.exit('{} already exists!'.format(ipsw_name))

    # TODO Add beta support CC @mcg29

    def getSignedVersions(self, beta=False):
        data = self.getDeviceJSONData()
        signed = list()
        for i in range(0, len(data['firmwares'])):
            version_from_json = data['firmwares'][i]['version']
            buildid_from_json = data['firmwares'][i]['buildid']
            signature_from_json = data['firmwares'][i]['signed']
            if signature_from_json:
                stuff = {
                    'iOS': version_from_json,
                    'buildid': buildid_from_json,
                    'signed': signature_from_json,
                    'filetype': 'ipsw'
                }
                if stuff not in signed:
                    signed.append(stuff)

        data = self.getDeviceJSONData('ota')
        for i in range(0, len(data['firmwares'])):
            version_from_json = data['firmwares'][i]['version']
            buildid_from_json = data['firmwares'][i]['buildid']
            signature_from_json = data['firmwares'][i]['signed']
            if signature_from_json:
                stuff = {
                    'iOS': version_from_json,
                    'buildid': buildid_from_json,
                    'signed': signature_from_json,
                    'filetype': 'ota'
                }
                if stuff not in signed:
                    signed.append(stuff)

        return signed

    def downloadFileFromArchive(self, path: str, filetype='ipsw'):
        url = self.getArchiveURL()
        if filetype == 'ipsw':
            with RemoteZip(url) as f:
                files = f.namelist()
                if not os.path.exists(path):
                    if path in files:
                        print('Downloading: {}'.format(path))
                        f.extract(path)
                else:
                    raise FileExistsError('{} already exists!'.format(path))

        elif filetype == 'ota':
            # TODO Add ota support
            pass

        else:
            raise ValueError('Only filetypes, "ipsw" and "ota" are supported!')

    # Same as above but we don't download, which does help with leftover json data and such

    def readFileFromArchive(self, path: str):
        url = self.getArchiveURL()
        with RemoteZip(url) as f:
            data = f.read(path)
            return data

    def getCodename(self):
        data = self.readFileFromArchive('BuildManifest.plist')
        stuff = plistlib.loads(data)
        codename = stuff['BuildIdentities'][0]['Info']['BuildTrain']
        return codename



import sys
from urllib.request import urlopen

try:
    from .ipswapi import API
    from .template import Template
except ImportError:
    raise


class iPhoneWiki(object):
    def __init__(self, device: str, version: str, buildid=False, codename=False, ota=False, beta=False):
        super().__init__()
        self.device = device
        self.version = version
        self.ota = ota
        self.beta = beta

        if buildid:
            self.buildid = buildid
        else:
            api = API(self.device, self.version)
            self.buildid = api.iOSToBuildid()

        if codename:
            self.codename = codename
        else:
            api = API(self.device, self.version)
            self.codename = api.getCodename()

        """
        Handles data on the iphonewiki page.

        Grabs key template from source code ;P
        """

    # TODO Make this check if the url is good. Probably have a dict saying like is got a response, or if there's
    # no keys for a particular device, iOS, etc.

    # Status codes:

    # 200, response was sent successfully

    # I need to check if there's even keys first

    def getWikiKeys(self):
        wikiUrl = 'https://www.theiphonewiki.com/w/index.php?title={}_{}_({})&action=edit'.format(
            self.codename, self.buildid, self.device)  # This is used for "ipsw" and "beta"
        response = urlopen(wikiUrl).read().decode('utf-8')
        if 'You can view and copy the source of this page.' not in response:
            sys.exit('Seems like there are no keys for: {} {} {}!'.format(
                self.device, self.version, self.buildid))
        else:
            template = Template()
            return template.parseTemplate(response)

    # TODO Maybe have it open an html file, the page to upload keys, importing template into the page, just needing to press "upload"

    def uploadWikiKeys(self):
        pass



import os
import plistlib
from shutil import rmtree
from urllib.request import Request, urlopen

try:
    from .ipswapi import API
    from .manifest import TSSManifest
except ImportError:
    raise

tss_url = 'https://gs.apple.com/TSS/controller?action=2'
tss_headers = {'User-Agent': 'InetURL/1.0', 'Proxy-Connection': 'Keep-Alive', 'Pragma': 'no-cache',
               'Content-type': 'text/xml; charset="utf-8"'}  # See https://www.theiphonewiki.com/wiki/SHSH_Protocol#Communication


class TSS(object):
    def __init__(self, device: str, ecid: str, version=False, apnonce='', sepnonce='', bbsnum='', useDFUCollidingNonces=False, shsh_path='shsh'):
        super().__init__()
        self.device = device
        self.ecid = ecid
        self.version = version
        self.apnonce = apnonce
        self.sepnonce = sepnonce
        self.bbsnum = bbsnum
        self.useDFUCollidingNonces = useDFUCollidingNonces
        self.shsh_path = shsh_path

    def makeTSSRequest(self, data: bytes):
        request = Request(tss_url, headers=tss_headers, data=data)

        response = urlopen(request, timeout=5.0)
        response_text = response.read().decode('utf-8')
        if 'STATUS=0' not in response_text:
            if response.headers['Content-Length'] == '0':
                print('Server returned no response... are you blacklisted?')
            else:
                print('Server error:', response_text)
        else:
            # Remove TSS response header
            return response_text[response_text.find('<?xml'):]

    def saveBlobs(self):
        api = API(self.device, self.version)
        signed_versions = api.getSignedVersions()

        if os.path.exists('.shsh'):
            rmtree('.shsh')
        os.mkdir('.shsh')

        if not os.path.exists(self.shsh_path):
            os.mkdir(self.shsh_path)

        tss_test_version_manifest = os.path.join(
            '.shsh', 'TSSTestVersionManifest.plist')

        manobj = TSSManifest()
        manobj.createTSSTestVersionManifest(tss_test_version_manifest)

        #  There is a better way to do the following (likely by using plistlib.loads/dumps)

        print('Sending TSS server version request...')

        with open(tss_test_version_manifest, 'rb') as v:
            tss_version_response = self.makeTSSRequest(v.read())

        with open(tss_test_version_manifest, 'w') as p:
            p.write(tss_version_response)

        with open(tss_test_version_manifest, 'rb') as v:
            tss_version_response = plistlib.load(v, fmt=plistlib.FMT_XML)

        tss_server_version = tss_version_response['@ServerVersion']
        if tss_server_version != '2.1.0':
            print('WARNING: TSS server version', tss_server_version,
                  'may not be supported by this script!')

        # FIXME Need to update the stuff below

        for stuff in signed_versions:
            version = stuff['iOS']
            buildid = stuff['buildid']

            # Copy the manifest inside which will contain the added string

            build_manifest = os.path.join(
                '.shsh', 'BuildManifest_{}_{}_{}.plist'.format(self.device, version, buildid))
            tss_manifest = os.path.join(
                '.shsh', 'TSSManifest_{}_{}_{}.plist'.format(self.device, version, buildid))

            api.downloadFileFromArchive('BuildManifest.plist')

            print('Converting from BuildManifest to TSS manifest...')
            apnonce = manobj.initFromBuildManifest(
                self.device, tss_manifest, build_manifest, self.ecid, apnonce=self.apnonce, sepnonce=self.sepnonce, bbsnum=self.bbsnum)['apnonce']

            print('Sending TSS request for', version, '(' + buildid + ')...')

            with open(tss_manifest, 'rb') as f:
                tss_response = self.makeTSSRequest(f.read())

            blob_path = os.path.join(self.shsh_path, '{}_{}_{}-{}_{}.shsh2'.format(
                self.ecid, self.device, version, buildid, apnonce))
            with open(blob_path, 'w+') as blob:
                blob.write(tss_response)
                print('Saved', version, 'blob to', blob_path)

        rmtree('.shsh')



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

REQUIREMENTS = [i.strip() for i in open("requirements.txt").readlines()]
# rbp https://stackoverflow.com/questions/6947988/when-to-use-pip-requirements-file-versus-install-requires-in-setup-py

setup(
    name="iospytools",
    version="1.0.4",
    author="Merculous",
    author_email="vycemerculous@gmail.com",
    description="provides useful tools/commands which are used in iOS research",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/merculous/ios-python-tools",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "iospytools = src.__main__:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    python_requires='>=3.5',
    install_requires=REQUIREMENTS
)



class IMG4(object):
    def __init__(self, file: str):
        super().__init__()

        self.file = file

        """

        Class to parse and interact with Apple's img4 formatted files.

        """

    def printTags(self):
        """
        Print the tags of an img3 file
        """

        with open(self.file, 'rb') as f:
            data = f.read()

            if b'IM4P' in data:
                print('I see a IM4P tag at index:', hex(data.find(b'IM4P')))

        f.close()
